@@ -5,7 +5,7 @@
 # Written 91-12-02 through 92-01-01 by Stephen McGee.
 # Modified 92-02-11 through 92-02-22 by Chris Arthur to further generalize.
 #
-# Copyright (C) 1991-2022 Free Software Foundation, Inc.
+# Copyright (C) 1991-2023 Free Software Foundation, Inc.
 # This file is part of GNU Make.
 #
 # GNU Make is free software; you can redistribute it and/or modify it under
@@ -43,6 +43,9 @@ $categories_passed = 0;
 $total_tests_run = 0;
 # The total number of individual tests that have passed
 $total_tests_passed = 0;
+# Set to true if something failed.  It could be that tests_run == tests_passed
+# even with failures, if we don't run tests for some reason.
+$some_test_failed = 0;
 # The number of tests in this category that have been run
 $tests_run = 0;
 # The number of tests in this category that have passed
@@ -57,15 +60,11 @@ $pathsep = undef;
 $test_passed = 1;
 
 # Timeout in seconds.  If the test takes longer than this we'll fail it.
-$test_timeout = 5;
-$test_timeout = 10 if $^O eq 'VMS';
+# This is to prevent hung tests.
+$test_timeout = 60;
 
 $diff_name = undef;
 
-# Create a temporary directory that tests can use, outside the temp
-# directory that make is using.
-$TEMPDIR = File::Temp->newdir();
-
 # Path to Perl
 $perl_name = $^X;
 if ($^O ne 'VMS') {
@@ -149,7 +148,7 @@ sub resetENV
   # through Perl 5.004.  It was fixed in Perl 5.004_01, but we don't
   # want to require that here, so just delete each one individually.
 
-  if ($^O ne 'VMS') {
+  if ($osname ne 'VMS') {
     foreach $v (keys %ENV) {
       delete $ENV{$v};
     }
@@ -201,11 +200,14 @@ sub toplevel
            'TZ', 'TMPDIR', 'HOME', 'USER', 'LOGNAME', 'PATH',
            'LD_LIBRARY_PATH',
            # *SAN things
-           'ASAN_OPTIONS', 'UBSAN_OPTIONS',
+           'ASAN_OPTIONS', 'UBSAN_OPTIONS', 'LSAN_OPTIONS',
            # Purify things
            'PURIFYOPTIONS',
            # Windows-specific things
            'Path', 'SystemRoot', 'TEMP', 'TMP', 'USERPROFILE', 'PATHEXT',
+           # z/OS specific things
+           'LIBPATH', '_BPXK_AUTOCVT',
+           '_TAG_REDIR_IN',  '_TAG_REDIR_OUT',
            # DJGPP-specific things
            'DJDIR', 'DJGPP', 'SHELL', 'COMSPEC', 'HOSTNAME', 'LFN',
            'FNCASE', '387', 'EMU387', 'GROUP'
@@ -368,6 +370,11 @@ sub toplevel
     print ($categories_failed == 1 ? "y" : "ies");
     print " Failed (See .$diffext* files in $workdir dir for details) :-(\n\n";
     return 0;
+  } elsif ($some_test_failed) {
+      # Something failed but no tests were marked failed... probably a syntax
+      # error in a test script
+    print "\nSome tests failed (See output for details) :-(\n\n";
+    return 0;
   }
 
   print "\n$total_tests_passed Test";
@@ -390,7 +397,7 @@ sub get_osname
   #
   # This is probably not specific enough.
   #
-  if ($osname =~ /MSWin32/i || $osname =~ /Windows/i
+  if ($osname =~ /MSWin32/i || $osname =~ /Windows/i || $osname =~ /msys/i
       || $osname =~ /MINGW32/i || $osname =~ /CYGWIN_NT/i) {
     $port_type = 'W32';
   }
@@ -413,8 +420,7 @@ sub get_osname
     $port_type = 'VMS-DCL';
   }
   # Everything else, right now, is UNIX.  Note that we should integrate
-  # the VOS support into this as well and get rid of $vos; we'll do
-  # that next time.
+  # the VOS support into this as well and get rid of $vos
   else {
     $port_type = 'UNIX';
   }
@@ -428,7 +434,7 @@ sub get_osname
   # See if the filesystem supports long file names with multiple
   # dots.  DOS doesn't.
   $short_filenames = 0;
-  (open (TOUCHFD, "> fancy.file.name") and close (TOUCHFD))
+  (open (TOUCHFD, '>', 'fancy.file.name') and close (TOUCHFD))
       or $short_filenames = 1;
   unlink ("fancy.file.name") or $short_filenames = 1;
 
@@ -611,7 +617,7 @@ sub run_all_tests
     $perl_testname = "$scriptpath$pathsep$testname";
     $testname =~ s/(\.pl|\.perl)$//;
     $testpath = "$workpath$pathsep$testname";
-    $extext = '_' if $^O eq 'VMS';
+    $extext = '_' if $osname eq 'VMS';
     $log_filename = "$testpath.$logext";
     $diff_filename = "$testpath.$diffext";
     $base_filename = "$testpath.$baseext";
@@ -650,6 +656,7 @@ sub run_all_tests
         warn "\n*** Couldn't parse $perl_testname\n";
       }
       $status = "FAILED ($tests_passed/$tests_run passed)";
+      $some_test_failed = 1;
 
     } elsif ($code == -1) {
       # Skipped... not supported
@@ -661,14 +668,17 @@ sub run_all_tests
       # the suite forgot to end with "1;".
       warn "\n*** Test returned $code\n";
       $status = "FAILED ($tests_passed/$tests_run passed)";
+      $some_test_failed = 1;
 
     } elsif ($tests_run == 0) {
       # Nothing was done!!
       $status = "FAILED (no tests found!)";
+      $some_test_failed = 1;
 
     } elsif ($tests_run > $tests_passed) {
       # Lose!
       $status = "FAILED ($tests_passed/$tests_run passed)";
+      $some_test_failed = 1;
 
     } else {
       # Win!
@@ -810,12 +820,164 @@ sub error
   die "$caller: $message";
 }
 
+sub compare_answer_vms
+{
+  my ($kgo, $log) = @_;
+
+  # VMS has extra blank lines in output sometimes.
+  # Ticket #41760
+  $log =~ s/\n\n+/\n/gm;
+  $log =~ s/\A\n+//g;
+  return 1 if ($kgo eq $log);
+
+  # VMS adding a "Waiting for unfinished jobs..."
+  # Remove it for now to see what else is going on.
+  $log =~ s/^.+\*\*\* Waiting for unfinished jobs.+$//m;
+  $log =~ s/\n\n/\n/gm;
+  $log =~ s/^\n+//gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS wants target device to exist or generates an error,
+  # Some test targets look like VMS devices and trip this.
+  $log =~ s/^.+\: no such device or address.*$//gim;
+  $log =~ s/\n\n/\n/gm;
+  $log =~ s/^\n+//gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS error message has a different case
+  $log =~ s/no such file /No such file /gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS is putting comas instead of spaces in output
+  $log =~ s/,/ /gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS Is sometimes adding extra leading spaces to output?
+  {
+     (my $mlog = $log) =~ s/^ +//gm;
+     return 1 if ($mlog eq $kgo);
+  }
+
+  # VMS port not handling POSIX encoded child status
+  # Translate error case it for now.
+  $log =~ s/0x1035a00a/1/gim;
+  return 1 if ($log =~ /\Q$kgo\E/i);
+
+  $log =~ s/0x1035a012/2/gim;
+  return 1 if ($log eq $kgo);
+
+  # Tests are using a UNIX null command, temp hack
+  # until this can be handled by the VMS port.
+  # ticket # 41761
+  $log =~ s/^.+DCL-W-NOCOMD.*$//gim;
+  $log =~ s/\n\n+/\n/gm;
+  $log =~ s/^\n+//gm;
+  return 1 if ($log eq $kgo);
+
+  # Tests are using exit 0;
+  # this generates a warning that should stop the make, but does not
+  $log =~ s/^.+NONAME-W-NOMSG.*$//gim;
+  $log =~ s/\n\n+/\n/gm;
+  $log =~ s/^\n+//gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS is sometimes adding single quotes to output?
+  $log =~ s/\'//gm;
+  return 1 if ($log eq $kgo);
+
+  # And missing an extra space in output
+  $kgo =~ s/\h\h+/ /gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS adding ; to end of some lines.
+  $log =~ s/;\n/\n/gm;
+  return 1 if ($log eq $kgo);
+
+  # VMS adding trailing space to end of some quoted lines.
+  $log =~ s/\h+\n/\n/gm;
+  return 1 if ($log eq $kgo);
+
+  # And VMS missing leading blank line
+  $kgo =~ s/\A\n//g;
+  return 1 if ($log eq $kgo);
+
+  # Unix double quotes showing up as single quotes on VMS.
+  $kgo =~ s/\"//g;
+  return 1 if ($log eq $kgo);
+
+  return 0;
+}
+
+sub convert_answer_zos
+{
+  my ($log) = @_;
+
+  # z/OS emits "Error 143" or "SIGTERM" instead of terminated
+  $log =~ s/Error 143/Terminated/gm;
+  $log =~ s/SIGTERM/Terminated/gm;
+
+  # z/OS error messages have a prefix
+  $log =~ s/EDC5129I No such file or directory\./No such file or directory/gm;
+  $log =~ s/FSUM7351 not found/not found/gm;
+
+  return $log;
+}
+
+sub compare_answer
+{
+  my ($kgo, $log) = @_;
+  my ($mkgo, $mlog);
+
+  # For make, get rid of any time skew error before comparing--too bad this
+  # has to go into the "generic" driver code :-/
+  $log =~ s/^.*modification time .*in the future.*\n//gm;
+  $log =~ s/^.*Clock skew detected.*\n//gm;
+  return 1 if ($log eq $kgo);
+
+  # Get rid of newline differences, forever
+  $kgo =~ s,\r\n,\n,gs;
+  $log =~ s,\r\n,\n,gs;
+  return 1 if ($log eq $kgo);
+
+  # Keep the originals in case it's a regex
+  $mkgo = $kgo;
+  $mlog = $log;
+
+  # z/OS has quirky outputs
+  if ($osname eq 'os390') {
+    $mlog = convert_answer_zos($mlog);
+    return 1 if ($mlog eq $kgo);
+  }
+
+  # Some versions of Perl on Windows use /c instead of C:
+  $mkgo =~ s,\b([A-Z]):,/\L$1,g;
+  $mlog =~ s,\b([A-Z]):,/\L$1,g;
+  return 1 if ($mlog eq $mkgo);
+
+  # See if it is a backslash problem (only on W32?)
+  $mkgo =~ tr,\\,/,;
+  $mlog =~ tr,\\,/,;
+  return 1 if ($mlog eq $mkgo);
+
+  # VMS is a whole thing...
+  return 1 if ($osname eq 'VMS' && compare_answer_vms($kgo, $log));
+
+  # See if the answer might be a regex.
+  if ($kgo =~ m,^/(.+)/$,) {
+    # Check the regex against both the original and modified strings
+    return 1 if ($log =~ /$1/);
+    return 1 if ($mlog =~ /$1/);
+  }
+
+  return 0;
+}
+
 my %old_tempfiles = ();
 
 sub compare_output
 {
   my ($answer, $logfile) = @_;
-  my ($slurp, $answer_matched, $extra) = ('', 0, 0);
+  my ($slurp, $matched, $extra) = ('', 0, 0);
 
   ++$tests_run;
 
@@ -835,168 +997,25 @@ sub compare_output
 
   if (! defined $answer) {
     print "Ignoring output ........ " if $debug;
-    $answer_matched = 1;
+    $matched = 1;
   } else {
     print "Comparing output ........ " if $debug;
 
-    $slurp = &read_file_into_string ($logfile);
-
-    # For make, get rid of any time skew error before comparing--too bad this
-    # has to go into the "generic" driver code :-/
-    $slurp =~ s/^.*modification time .*in the future.*\n//gm;
-    $slurp =~ s/^.*Clock skew detected.*\n//gm;
-
-    if ($slurp eq $answer) {
-        $answer_matched = 1;
-    } else {
-      # See if it is a slash or CRLF problem
-      my ($answer_mod, $slurp_mod) = ($answer, $slurp);
-
-      $answer_mod =~ tr,\\,/,;
-      $answer_mod =~ s,\r\n,\n,gs;
-
-      $slurp_mod =~ tr,\\,/,;
-      $slurp_mod =~ s,\r\n,\n,gs;
-
-      $answer_matched = ($slurp_mod eq $answer_mod);
-      if ($^O eq 'VMS') {
-
-        # VMS has extra blank lines in output sometimes.
-        # Ticket #41760
-        if (!$answer_matched) {
-          $slurp_mod =~ s/\n\n+/\n/gm;
-          $slurp_mod =~ s/\A\n+//g;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS adding a "Waiting for unfinished jobs..."
-        # Remove it for now to see what else is going on.
-        if (!$answer_matched) {
-          $slurp_mod =~ s/^.+\*\*\* Waiting for unfinished jobs.+$//m;
-          $slurp_mod =~ s/\n\n/\n/gm;
-          $slurp_mod =~ s/^\n+//gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS wants target device to exist or generates an error,
-        # Some test tagets look like VMS devices and trip this.
-        if (!$answer_matched) {
-          $slurp_mod =~ s/^.+\: no such device or address.*$//gim;
-          $slurp_mod =~ s/\n\n/\n/gm;
-          $slurp_mod =~ s/^\n+//gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS error message has a different case
-        if (!$answer_matched) {
-          $slurp_mod =~ s/no such file /No such file /gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS is putting comas instead of spaces in output
-        if (!$answer_matched) {
-          $slurp_mod =~ s/,/ /gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS Is sometimes adding extra leading spaces to output?
-        if (!$answer_matched) {
-           my $slurp_mod = $slurp_mod;
-           $slurp_mod =~ s/^ +//gm;
-           $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS port not handling POSIX encoded child status
-        # Translate error case it for now.
-        if (!$answer_matched) {
-          $slurp_mod =~ s/0x1035a00a/1/gim;
-          $answer_matched = 1 if $slurp_mod =~ /\Q$answer_mod\E/i;
-
-        }
-        if (!$answer_matched) {
-          $slurp_mod =~ s/0x1035a012/2/gim;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # Tests are using a UNIX null command, temp hack
-        # until this can be handled by the VMS port.
-        # ticket # 41761
-        if (!$answer_matched) {
-          $slurp_mod =~ s/^.+DCL-W-NOCOMD.*$//gim;
-          $slurp_mod =~ s/\n\n+/\n/gm;
-          $slurp_mod =~ s/^\n+//gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-        # Tests are using exit 0;
-        # this generates a warning that should stop the make, but does not
-        if (!$answer_matched) {
-          $slurp_mod =~ s/^.+NONAME-W-NOMSG.*$//gim;
-          $slurp_mod =~ s/\n\n+/\n/gm;
-          $slurp_mod =~ s/^\n+//gm;
-          $answer_matched = ($slurp_mod eq $answer_mod);
-        }
-
-        # VMS is sometimes adding single quotes to output?
-        if (!$answer_matched) {
-          my $noq_slurp_mod = $slurp_mod;
-          $noq_slurp_mod =~ s/\'//gm;
-          $answer_matched = ($noq_slurp_mod eq $answer_mod);
-
-          # And missing an extra space in output
-          if (!$answer_matched) {
-            $noq_answer_mod = $answer_mod;
-            $noq_answer_mod =~ s/\h\h+/ /gm;
-            $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
-          }
-
-          # VMS adding ; to end of some lines.
-          if (!$answer_matched) {
-            $noq_slurp_mod =~ s/;\n/\n/gm;
-            $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
-          }
-
-          # VMS adding trailing space to end of some quoted lines.
-          if (!$answer_matched) {
-            $noq_slurp_mod =~ s/\h+\n/\n/gm;
-            $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
-          }
-
-          # And VMS missing leading blank line
-          if (!$answer_matched) {
-            $noq_answer_mod =~ s/\A\n//g;
-            $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
-          }
-
-          # Unix double quotes showing up as single quotes on VMS.
-          if (!$answer_matched) {
-            $noq_answer_mod =~ s/\"//g;
-            $answer_matched = ($noq_slurp_mod eq $noq_answer_mod);
-          }
-        }
-      }
-
-      # If it still doesn't match, see if the answer might be a regex.
-      if (!$answer_matched && $answer =~ m,^/(.+)/$,) {
-        $answer_matched = ($slurp =~ /$1/);
-        if (!$answer_matched && $answer_mod =~ m,^/(.+)/$,) {
-            $answer_matched = ($slurp_mod =~ /$1/);
-        }
-      }
-    }
+    $matched = compare_answer($answer, &read_file_into_string ($logfile));
   }
 
-  if ($keep || ! $answer_matched) {
+  if ($keep || ! $matched) {
     &create_file(&get_basefile, $answer);
     &create_file(&get_runfile, $command_string);
   }
 
-  if ($answer_matched && $test_passed && !$extra) {
+  if ($matched && $test_passed && !$extra) {
     print "ok\n" if $debug;
     ++$tests_passed;
     return 1;
   }
 
-  if (! $answer_matched) {
+  if (! $matched) {
     print "DIFFERENT OUTPUT\n" if $debug;
 
     print "\nCreating Difference File ...\n" if $debug;
@@ -1004,10 +1023,11 @@ sub compare_output
     # Create the difference file
     my $base = get_basefile();
     if ($diff_name) {
-        my $command = "$diff_name -c $base $logfile";
-        &run_command_with_output(get_difffile(), $command);
+        &run_command_with_output(get_difffile(),
+                                 "$diff_name -c $base $logfile");
     } else {
-        create_file(get_difffile(), "Log file $logfile differs from base file $base\n");
+        create_file(get_difffile(),
+                    "Log file $logfile differs from base file $base\n");
     }
   }
 
@@ -1070,7 +1090,7 @@ sub detach_default_output
   @OUTSTACK or error("default output stack has flown under!\n", 1);
 
   close(STDOUT);
-  close(STDERR) unless $^O eq 'VMS';
+  close(STDERR) unless $osname eq 'VMS';
 
 
   open (STDOUT, '>&', pop @OUTSTACK) or error("ddo: $! duping STDOUT\n", 1);
@@ -1080,7 +1100,7 @@ sub detach_default_output
 sub _run_with_timeout
 {
   my $code;
-  if ($^O eq 'VMS') {
+  if ($osname eq 'VMS') {
     #local $SIG{ALRM} = sub {
     #    my $e = $ERRSTACK[0];
     #    print $e "\nTest timed out after $test_timeout seconds\n";
@@ -1106,9 +1126,10 @@ sub _run_with_timeout
       $code = (($vms_code & 0xFFF) >> 3) * 256;
     }
 
-  } elsif ($port_type eq 'W32') {
+  } elsif ($port_type eq 'W32' && $^O ne 'msys') {
+    # Using ActiveState Perl (?)
     my $pid = system(1, @_);
-    $pid > 0 or die "Cannot execute $_[0]\n";
+    $pid > 0 or die "Cannot execute $_[0]: $!\n";
     local $SIG{ALRM} = sub {
       my $e = $ERRSTACK[0];
       print $e "\nTest timed out after $test_timeout seconds\n";
@@ -1175,7 +1196,7 @@ sub run_command
   print "\nrun_command: @_\n" if $debug;
   my $code = _run_command(@_);
   print "run_command returned $code.\n" if $debug;
-  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $^O eq 'VMS';
+  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $osname eq 'VMS';
   return $code;
 }
 
@@ -1197,7 +1218,7 @@ sub run_command_with_output
   $err and die $err;
 
   print "run_command_with_output returned $code.\n" if $debug;
-  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $^O eq 'VMS';
+  print "vms status = ${^CHILD_ERROR_NATIVE}\n" if $debug and $osname eq 'VMS';
   return $code;
 }
 
@@ -1245,7 +1266,7 @@ sub remove_directory_tree_inner
         return 0;
       }
     } else {
-      if ($^O ne 'VMS') {
+      if ($osname ne 'VMS') {
         if (!unlink $object) {
           print "Cannot unlink $object: $!\n";
           return 0;
@@ -1269,7 +1290,7 @@ sub remove_directory_tree_inner
 #
 #  foreach my $file (@filenames) {
 #    utime ($now, $now, $file)
-#          or (open (TOUCHFD, ">> $file") and close (TOUCHFD))
+#          or (open (TOUCHFD, '>>', $file) and close (TOUCHFD))
 #               or &error ("Couldn't touch $file: $!\n", 1);
 #  }
 #  return 1;
@@ -1316,7 +1337,7 @@ sub create_file
 {
   my ($filename, @lines) = @_;
 
-  open (CF, "> $filename") or &error ("Couldn't open $filename: $!\n", 1);
+  open (CF, '>', $filename) or &error ("Couldn't open '$filename': $!\n", 1);
   foreach $line (@lines) {
     print CF $line;
   }
